package Async::Template::Provider;

#! @file
#! @author: Serguei Okladnikov <oklaspec@gmail.com>
#! @date 21.05.2015

use base 'Template::Provider';

use Async::Template::Document;

use constant DOCUMENT => 'Async::Template::Document';

1;
